//
// Created by illyum on 8/15/2024.
//

#include "dmx.h"
#include <iostream>
#include <chrono>
#include <thread>
#include "windows.h"

DMXEngine::DMXEngine(const char* portName, int baudRate, int dataSize)
        : running(false), buffer(dataSize, 0) {
    hSerial = CreateFileA(portName, GENERIC_WRITE, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if (hSerial == INVALID_HANDLE_VALUE) {
        std::cerr << "Error opening " << portName << " port" << std::endl;
        exit(EXIT_FAILURE);
    }

    configurePort();
}

DMXEngine::~DMXEngine() {
    if (running) {
        stop();
    }
    CloseHandle(hSerial);
}

void DMXEngine::configurePort() {
    DCB dcbSerialParams = { 0 };
    dcbSerialParams.DCBlength = sizeof(dcbSerialParams);
    if (!GetCommState(hSerial, &dcbSerialParams)) {
        std::cerr << "Error getting COM port state" << std::endl;
        CloseHandle(hSerial);
        exit(EXIT_FAILURE);
    }

    dcbSerialParams.BaudRate = 250000;
    dcbSerialParams.ByteSize = 8;
    dcbSerialParams.StopBits = TWOSTOPBITS;
    dcbSerialParams.Parity = NOPARITY;

    if (!SetCommState(hSerial, &dcbSerialParams)) {
        std::cerr << "Error setting COM port state" << std::endl;
        CloseHandle(hSerial);
        exit(EXIT_FAILURE);
    }

    COMMTIMEOUTS timeouts = { 0 };
    timeouts.ReadIntervalTimeout = 50;
    timeouts.ReadTotalTimeoutConstant = 50;
    timeouts.ReadTotalTimeoutMultiplier = 10;
    timeouts.WriteTotalTimeoutConstant = 50;
    timeouts.WriteTotalTimeoutMultiplier = 10;

    if (!SetCommTimeouts(hSerial, &timeouts)) {
        std::cerr << "Error setting COM port timeouts" << std::endl;
        CloseHandle(hSerial);
        exit(EXIT_FAILURE);
    }
}

void DMXEngine::start() {
    running = true;
    dmxThread = std::thread(dmxThreadFunc, this);
}

void DMXEngine::stop() {
    running = false;
    if (dmxThread.joinable()) {
        dmxThread.join();
    }

    DMXEngine::dumpZeros();
}


void DMXEngine::setChannelValue(int channel, unsigned char value) {
    if (channel >= 0 && channel < buffer.size()) {
        buffer[channel] = value;
    }
}

void DMXEngine::sendDMXData() {
    while (running) {
        // Break condition
        EscapeCommFunction(hSerial, SETBREAK);
        std::this_thread::sleep_for(std::chrono::milliseconds(1));
        EscapeCommFunction(hSerial, CLRBREAK);
        std::this_thread::sleep_for(std::chrono::milliseconds(1)); // MAB

        DWORD bytesWritten;
        unsigned char startCode = 0;
        WriteFile(hSerial, &startCode, 1, &bytesWritten, NULL);
        WriteFile(hSerial, buffer.data(), buffer.size(), &bytesWritten, NULL);

        std::this_thread::sleep_for(std::chrono::milliseconds(25));
    }
}

void DMXEngine::dumpZeros() {
    std::fill(buffer.begin(), buffer.end(), 0);

    EscapeCommFunction(hSerial, SETBREAK);
    std::this_thread::sleep_for(std::chrono::milliseconds(1));
    EscapeCommFunction(hSerial, CLRBREAK);
    std::this_thread::sleep_for(std::chrono::milliseconds(1)); // MAB

    DWORD bytesWritten;
    unsigned char startCode = 0;
    WriteFile(hSerial, &startCode, 1, &bytesWritten, NULL);
    WriteFile(hSerial, buffer.data(), buffer.size(), &bytesWritten, NULL);
}

void DMXEngine::dmxThreadFunc(DMXEngine* engine) {
    engine->sendDMXData();
}
